import daj.*;
import java.util.Vector;

/**
 *	Algorithms for Distributed Systems, SS 00, Project)
 *
 *	AsynchBFS algorithm program.
 *	Derived from Dijkstra-Scholten algorithm program class
 *	to implement monitoring of termination detection.
 *
 *	@author Jrgen Hartl, 9756179 / 881
 *	@author Rene Kobler, 9655888 / 881
 *	@author Christian Gintenreiter, 9655980 / 881
 *	@version 1.0
 */
class AsynchBFSProgram extends DijkstraScholtenProgram
{
	/**
	 *	Constant defines infinite distance from root value.
	 */
	protected static final int INFINITE = -1;

	/**
	 *	Constant defines no parent value.
	 */
	protected static final int NOPARENT = -1;

	/**
	 *	Holds parent channel index (algorithm state variable).
	 */
	protected int parent;

	/**
	 *	Holds distance to root (algorithm state variable).
	 */
	protected int dist;

	/**
	 *	Indicates that node is quiescent.
	 */
	protected boolean quiescent;

	/**
	 *	Holds AsynchBFS message value FIFO send queues for each channel.
	 */
	protected Vector[] sendQueue;

	/**
	 *	Holds parent node description.
	 */
	protected String parentId;

	/**
	 *	Constructor.
	 *
	 *	@param number Index of program in program array.
	 */
	public AsynchBFSProgram(int number)
	{
		super(number);
	}

	/**
	 *	Initializes state variables as defined for AsynchBFS algorithm.
	 */
	public void init()
	{
		int i;

		super.init();

		if (in().getSize() != out().getSize())
		{
			System.err.println("Invalid network");
			exit(-1);
		}
		parentId = new String("(unknown)");

		parent = NOPARENT;
		dist = -1;
		sendQueue = new Vector[in().getSize()];
		for (i = 0; i < sendQueue.length; i++)
		{
			sendQueue[i] = new Vector();
		}

		quiescent = true;
	}

	/**
	 *	AsynchBFS algorithm execution. Dijkstra-Scholten monitoring calls
	 *	are inserted.
	 */
	public void main()
	{
		InChannelSet msgIn = in();
		OutChannelSet msgOut = out();
		GlobalAssertion dijkstraScholtenLemma = new DijkstraScholtenAssertion();
		AsynchBFSMessage msg;
		boolean terminationDetected;
		int channelCnt = msgIn.getSize();
		int channel;
		int msgDist;
		int i;

		init();

		assert(dijkstraScholtenLemma); // valid before input event

		if (number == 0)
		{
			dist = 0;
			for (i = 0; i < sendQueue.length; i++)
			{
				sendQueue[i].addElement(new Integer(0));
			}
			setSource();
			quiescent = false;
		}

		while (true)
		{
			assert(dijkstraScholtenLemma);
			if (super.isDone()) // termination detected
			{
				break;
			}
			assert(dijkstraScholtenLemma);

			channel = msgIn.select(1);
			if (channel != -1)
			{
				msg = (AsynchBFSMessage)super.receive(msgIn.getChannel(channel), channel);
				if (msg != null)
				{
					quiescent = false;
					msgDist = msg.getValue();
					if ((dist == INFINITE) || ((msgDist != INFINITE) && (msgDist + 1 < dist)))
					{
						dist = msgDist + 1;
						parent = channel;
						parentId = msg.getSenderId();
						for (i = 0; i < sendQueue.length; i++)
						{
							if (i != channel)
							{
								sendQueue[i].addElement(new Integer(dist));
							}
						}
					}
				}
				continue;
			}

			if (super.cleanUp())
			{
				continue;
			}

			channel = Math.abs(Main.random.nextInt()) % sendQueue.length;
			for (i = 0; (i < sendQueue.length) && sendQueue[channel].isEmpty(); i++)
			{
				channel = (channel + 1) % channelCnt;
			}
			if (!sendQueue[channel].isEmpty())
			{
				quiescent = false;
				super.send(
					msgOut.getChannel(channel),
					new AsynchBFSMessage(((Integer)sendQueue[channel].firstElement()).intValue())
				);
				sendQueue[channel].removeElementAt(0);
				super.postprocessSend(channel);
				continue;
			}
			if (super.sendAcknowledgement())
			{
				continue;
			}
			quiescent = true;
		}
		assert(dijkstraScholtenLemma);
		interrupt();
	}

	/**
	 *	Indicates that node is quiescent.
	 *
	 *	@return True if node is quiescent.
	 */
	protected boolean isQuiescent()
	{
		return quiescent;
	}

	/**
	 *	Returns description of the node.
	 *
	 *	@return Description of the node.
	 */
	public String getText()
	{
		StringBuffer text = new StringBuffer("AsynchBFS program:\nparent channel: ");

		text.append(parent);
		text.append("\nparent id: ");
		text.append(parentId);
		text.append("\ncurrent distance to root: ");
		if (dist == INFINITE)
		{
			text.append("infinite");
		}
		else
		{
			text.append(dist);
		}
		text.append("\n");
		text.append(super.getText());
		text.append("\nquiescent: ");
		text.append(quiescent);
		return text.toString();
	}
}
