import daj.*;

/**
 *	Algorithms for Distributed Systems, SS 00, Project)
 *
 *	Dijkstra-Scholten algorithm assertion (lemma 19.1, page 624).
 *
 *	@author Jrgen Hartl, 9756179 / 881
 *	@author Rene Kobler, 9655888 / 881
 *	@author Christian Gintenreiter, 9655980 / 881
 *	@version 1.0
 */
class DijkstraScholtenAssertion extends GlobalAssertion
{
	/**
	 *	Returns message displayed if assertion fails.
	 *
	 *	@return Message displayed if assertion fails.
	 */
	public String getText()
	{
		return "Dijkstra-Scholten assertion (Lemma 19.1, page 624) failed";
	}

	/**
	 * Checks point 1. of lemma 19.1 on page 624
	 *
	 * @param prog Array of Dijkstra-Scholten programs.
	 * @param inputNode Node that received initial input.
	 * @return True if lemma 19.1, point 1. holds.
	 */
	public boolean check1(DijkstraScholtenProgram prog[], int inputNode)
	{
		return
			(
				(prog[inputNode].status == DijkstraScholtenProgram.IDLE) ||
				(prog[inputNode].status == DijkstraScholtenProgram.SOURCE)
			) &&
			(prog[inputNode].parent == DijkstraScholtenProgram.NOPARENT);
	}

	/**
	 * Checks point 2. of lemma 19.1 on page 624
	 *
	 * @param prog Array of Dijkstra-Scholten programs.
	 * @param inputNode Node that received initial input.
	 * @return True if lemma 19.1, point 2. holds.
	 */
	public boolean check2(DijkstraScholtenProgram prog[], int inputNode)
	{
		boolean valid = true;
		int i;

		for (i = 0; valid && (i < prog.length); i++)
		{
			if (i != inputNode)
			{
				valid =
					(prog[i].status == DijkstraScholtenProgram.IDLE) ||
					(prog[i].status == DijkstraScholtenProgram.NONSOURCE);
				if (valid && (prog[i].status == DijkstraScholtenProgram.NONSOURCE))
				{
					valid = prog[i].parent != DijkstraScholtenProgram.NOPARENT;
				}
			}
		}
		return valid;
	}

	/**
	 * Checks point 3. of lemma 19.1 on page 624
	 *
	 * @param prog Array of Dijkstra-Scholten programs.
	 * @return True if lemma 19.1, point 3. holds.
	 */
	public boolean check3(DijkstraScholtenProgram prog[])
	{
		boolean valid = true;
		int i;

		for (i = 0; valid && (i < prog.length); i++)
		{
			if (prog[i].status == DijkstraScholtenProgram.IDLE)
			{
				valid =
					(prog[i].parent == DijkstraScholtenProgram.NOPARENT) &&
					prog[i].isQuiescent() && prog[i].noDeficit();
			}
		}
		return valid;
	}

	/**
	 * Checks point 4. of lemma 19.1 on page 624
	 *
	 * @param prog Array of Dijkstra-Scholten programs.
	 * @return True if lemma 19.1, point 4. holds.
	 */
	public boolean check4(DijkstraScholtenProgram prog[])
	{
		boolean valid = true;
		DijkstraScholtenMessage msg;
		Message[] messages;
		int partner;
		int channel;
		int cnt;
		int i, j, k;

		for (i = 0; valid && (i < prog.length); i++)
		{
			for (j = 0; valid && (j < prog[i].deficit.length); j++)
			{
				partner = prog[i].getProgram(j);
				if (partner < 0)
				{
					valid = (prog[i].deficit[j] == 0);
				}
				else
				{
					channel = prog[partner].getChannel(i);
					if (channel < 0)
					{
						System.err.println("Invalid network");
						return false;
					}
					cnt = prog[partner].sendAcknowledgement[channel];
					messages = getMessages(prog[i].out(j));
					for (k = 0; k < messages.length; k++)
					{
						if (((DijkstraScholtenMessage)messages[k]).isAlgorithmMessage())
						{
							cnt++;
						}
					}
					messages = getMessages(prog[i].in(j));
					for (k = 0; k < messages.length; k++)
					{
						if (((DijkstraScholtenMessage)messages[k]).isAcknowledgement())
						{
							cnt++;
						}
					}
					if (prog[partner].parent == channel)
					{
						cnt++;
					}
					valid = (prog[i].deficit[j] == cnt);
				}
			}
		}
		return valid;
	}

	/**
	 * Checks point 5. of lemma 19.1 on page 624
	 *
	 * @param prog Array of Dijkstra-Scholten programs.
	 * @param inputNode Node that received initial input.
	 * @return True if lemma 19.1, point 5. holds.
	 */
	public boolean check5(DijkstraScholtenProgram prog[], int inputNode)
	{
		boolean valid = true;
		boolean continueSearch;
		boolean[] visited;
		int i;

		if (prog[inputNode].status == DijkstraScholtenProgram.SOURCE)
		{
			visited = new boolean[prog.length];
			for (i = 0; i < visited.length; i++)
			{
				visited[i] = false;
			}
			visited[inputNode] = true;
			valid = (prog[inputNode].parent == DijkstraScholtenProgram.NOPARENT);
			continueSearch = valid;

			while (continueSearch)
			{
				continueSearch = false;
				for (i = 0; i < prog.length; i++)
				{
					if ((!visited[i]) && (prog[i].parent != DijkstraScholtenProgram.NOPARENT) && visited[prog[i].getProgram(prog[i].parent)])
					{
						visited[i] = true;
						continueSearch = true;
					}
				}
			}
			for (i = 0; valid && (i < prog.length); i++)
			{
				valid =
					(
						visited[i] && (
							(prog[i].parent != DijkstraScholtenProgram.NOPARENT) || (
								(prog[i].parent == DijkstraScholtenProgram.NOPARENT) &&
								(i == inputNode)
							)
						) &&
						(prog[i].status != DijkstraScholtenProgram.IDLE)
					) ||
					(
						(!visited[i]) && (prog[i].parent == DijkstraScholtenProgram.NOPARENT) &&
						(prog[i].status == DijkstraScholtenProgram.IDLE)
					);
			}
		}
		return valid;
	}

	/**
	 * Checks point 6. of lemma 19.1 on page 624
	 *
	 * @param prog Array of Dijkstra-Scholten programs.
	 * @param inputNode Node that received initial input.
	 * @return True if lemma 19.1, point 6. holds.
	 */
	public boolean check6(DijkstraScholtenProgram prog[], int inputNode)
	{
		boolean valid = true;
		int i, j;

		if (prog[inputNode].status == DijkstraScholtenProgram.IDLE)
		{
			for (i = 0; valid && (i < prog.length); i++)
			{
				valid = (prog[i].status == DijkstraScholtenProgram.IDLE);
				for (j = 0; valid && (j < prog[i].in().getSize()); j++)
				{
					valid = (getMessages(prog[i].in(j)).length == 0);
				}
				for (j = 0; valid && (j < prog[i].out().getSize()); j++)
				{
					valid = (getMessages(prog[i].out(j)).length == 0);
				}
			}
		}
		return valid;
	}

	/**
	 * Checks assertion for given Dijkstra-Scholten program array.
	 *
	 * @param prog Array of Dijkstra-Scholten programs.
	 * @return True if lemma 19.1, holds.
	 */
	public boolean assert(Program prog[])
	{
		DijkstraScholtenProgram[] dsprog = new DijkstraScholtenProgram[prog.length];
		int inputNode = -1;
		int i;

		for (i = 0; i < dsprog.length; i++)
		{
			dsprog[i] = (DijkstraScholtenProgram)prog[i];
			if (dsprog[i].isSource)
			{
				if (inputNode < 0)
				{
					inputNode = i;
				}
				else
				{
					return false;
				}
			}
		}

		if (inputNode < 0)
		{
			return true;
		}
		else
		{
			return
				check1(dsprog, inputNode) &&
				check2(dsprog, inputNode) &&
				check3(dsprog) &&
				check4(dsprog) &&
				check5(dsprog, inputNode) &&
				check6(dsprog, inputNode);
		}
	}
}
