import daj.*;

/**
 *	Algorithms for Distributed Systems, SS 00, Project)
 *
 *	Dijkstra-Scholten algorithm program.
 *
 *	@author Jrgen Hartl, 9756179 / 881
 *	@author Rene Kobler, 9655888 / 881
 *	@author Christian Gintenreiter, 9655980 / 881
 *	@version 1.0
 */
abstract class DijkstraScholtenProgram extends Program
{
	/**
	 *	Constant defines no parent value.
	 */
	protected static final int NOPARENT = -1;

	/**
	 *	Constant defines idle state.
	 */
	protected static final int IDLE = 0;

	/**
	 *	Constant defines source state.
	 */
	protected static final int SOURCE = 1;

	/**
	 *	Constant defines non-source state.
	 */
	protected static final int NONSOURCE = 2;

	/**
	 *	Constant defines state descriptions.
	 */
	protected static final String[] STATUSDESCRIPTION = {
		new String("idle"),
		new String("source"),
		new String("nonsource")
	};

	/**
	 *	Holds mapping from program index to channel index.
	 */
	protected int[] channelMatrix;

	/**
	 *	Holds mapping from channel index to program index.
	 */
	protected int[] programMatrix;

	/**
	 *	Indicates that node is source node that received initial input.
	 */
	protected boolean isSource;

	/**
	 *	Holds index in program array.
	 */
	protected int number;

	/**
	 *	Holds node description.
	 */
	protected String nodeId;

	/**
	 *	Holds parent node description.
	 */
	protected String parentId;

	/**
	 *	Holds parent node channel index (algorithm state variable).
	 */
	protected int parent;

	/**
	 *	Holds node status (algorithm state variable).
	 */
	protected int status;

	/**
	 *	Holds Dijkstra-Scholten acknowledgement message
	 *	FIFO send queues for each channel, realized as
	 *	message counter (algorithm state variable).
	 */
	protected int[] sendAcknowledgement;

	/**
	 *	Holds acknowledgment deficit counter for each
	 *	channel (algorithm state variable).
	 */
	protected int[] deficit;

	/**
	 *	Constructor.
	 *
	 *	@param number Index of program in program array.
	 */
	public DijkstraScholtenProgram(int number)
	{
		this.number = number;
		nodeId = new String("(unknown)");
		isSource = false;
	}

	/**
	 *	Sets node description.
	 *
	 *	@param nodeId Node description.
	 */
	public void setNodeId(String nodeId)
	{
		this.nodeId = nodeId;
	}

	/**
	 *	Initializes connection info variables.
	 *
	 *	@param channelCnt Number of unidirectional channels (in = out) of this node.
	 *	@param nodeCnt Number of nodes in network.
	 */
	public void initConnectionInfo(int channelCnt, int nodeCnt)
	{
		int i;

		programMatrix = new int[channelCnt];
		channelMatrix = new int[nodeCnt];

		for (i = 0; i < programMatrix.length; i++)
		{
			programMatrix[i] = -1;
		}
		for (i = 0; i < channelMatrix.length; i++)
		{
			channelMatrix[i] = -1;
		}
	}

	/**
	 *	Sets connection info variables.
	 *
	 *	@param channel Unidirectional channel index.
	 *	@param program Index of given channel's partner program index in program array.
	 */
	public void setConnectionInfo(int channel, int program)
	{
		programMatrix[channel] = program;
		channelMatrix[program] = channel;
	}

	/**
	 *	Returns the local unidirectional channel index for given partner program index.
	 *
	 *	@param program Partner program index.
	 *	@return The local unidirectional channel index for given partner program index.
	 */
	public int getChannel(int program)
	{
		return channelMatrix[program];
	}

	/**
	 *	Returns the partner program index for given local unidirectional channel index .
	 *
	 *	@param channel Local unidirectional channel index.
	 *	@return The partner program index for given local unidirectional channel index ..
	 */
	public int getProgram(int channel)
	{
		return programMatrix[channel];
	}

	/**
	 *	Initializes state variables as defined for Dijkstra-Scholten algorithm.
	 */
	protected void init()
	{
		int channelCnt = in().getSize();
		int i;

		if (channelCnt != out().getSize())
		{
			System.err.println("Invalid network");
			exit(-1);
		}
		parentId = new String("(unknown)");
		isSource = false;

		status = IDLE;
		parent = NOPARENT;
		sendAcknowledgement = new int[channelCnt];
		deficit = new int[channelCnt];
		for (i = 0; i < sendAcknowledgement.length; i++)
		{
			sendAcknowledgement[i] = 0;
			deficit[i] = 0;
		}
	}

	/**
	 *	Sets the node to be the source node which received the initial input.
	 */
	public void setSource()
	{
		status = SOURCE;
		isSource = true;
	}

	/**
	 *	Wraps channel receive transitions to filter Dijkstra-Scholten
	 *	acknowledgement messages. Given channel must contain a message to be read.
	 *
	 *	@param channel Channel message is read from.
	 *	@param channelIndex Index of channel in the nodes in-channel set.
	 *	@return The monitored algorithm message received or null if
	 *		Dijkstra-Scholten acknowledgement message was received.
	 */
	protected DijkstraScholtenMessage receive(InChannel channel, int channelIndex)
	{
		DijkstraScholtenMessage msg = (DijkstraScholtenMessage)channel.receive();

		if (msg.isAcknowledgement())
		{
			deficit[channelIndex]--;
			return null;
		}
		else
		{
			if (status == IDLE)
			{
				status = NONSOURCE;
				parent = channelIndex;

				parentId = msg.getSenderId();
			}
			else
			{
				sendAcknowledgement[channelIndex]++;
			}
			return msg;
		}
	}

	/**
	 *	Wraps channel send to set sender node id.
	 *
	 *	@param channel Channel message is sent to.
	 *	@param msg Message to be sent.
	 */
	protected void send(OutChannel channel, DijkstraScholtenMessage msg)
	{
		msg.setSenderId(nodeId);
		channel.send(msg);
	}

	/**
	 *	Realizes effects of successful monitored algorithm send transition.
	 *
	 *	@param channelIndex Index of channel in the nodes out-channel set.
	 */
	protected void postprocessSend(int channelIndex)
	{
		deficit[channelIndex]++;
	}

	/**
	 *	Implements output transition sendAcknowledgement of Dijkstra-Scholten
	 *	algorithm.
	 *
	 *	@return True if transition preconditions hold and effects took place.
	 */
	protected boolean sendAcknowledgement()
	{
		int channel = Math.abs(Main.random.nextInt()) % sendAcknowledgement.length;
		int i;

		for (i = 0; (i < sendAcknowledgement.length) && (sendAcknowledgement[channel] <= 0); i++)
		{
			channel = (channel + 1) % sendAcknowledgement.length;
		}
		if (sendAcknowledgement[channel] > 0)
		{
			out(channel).send(new DijkstraScholtenMessage(DijkstraScholtenMessage.ACKNOWLEDGEMENT, nodeId));
			sendAcknowledgement[channel]--;
			return true;
		}
		return false;
	}

	/**
	 *	Implements output transition done of Dijkstra-Scholten algorithm.
	 *
	 *	@return True if transition preconditions hold and effects took place.
	 */
	public boolean isDone()
	{
		if ((status == SOURCE) && isQuiescent() && noDeficit())
		{
			status = IDLE;
			return true;
		}
		return false;
	}

	/**
	 *	Implements internal transition cleanUp of Dijkstra-Scholten algorithm.
	 *
	 *	@return True if transition preconditions hold and effects took place.
	 */
	protected boolean cleanUp()
	{
		if ((status == NONSOURCE) && isQuiescent() && noDeficit())
		{
			sendAcknowledgement[parent]++;
			status = IDLE;
			parent = NOPARENT;

			parentId = new String("(unknown)");
			return true;
		}
		return false;
	}

	/**
	 *	Returns true if there are no deficits for all channels.
	 *
	 *	@return True if there are no deficits for all channels.
	 */
	protected boolean noDeficit()
	{
		boolean noDeficit = true;
		int i = 0;

		while ((i < deficit.length) && noDeficit)
		{
			noDeficit = (deficit[i] == 0);
			i++;
		}
		return noDeficit;
	}

	/**
	 *	Indicates that node is quiescent.
	 *
	 *	@return True if node is quiescent.
	 */
	protected abstract boolean isQuiescent();

	/**
	 *	Returns description of the node.
	 *
	 *	@return Description of the node.
	 */
	public String getText()
	{
		StringBuffer text = new StringBuffer("Dijkstra-Scholten program:\nnode number: ");
		int i;

		text.append(number);
		text.append("\nnode id: ");
		text.append(nodeId);
		text.append("\nstatus: ");
		text.append(STATUSDESCRIPTION[status]);
		text.append("\nparent channel: ");
		text.append(parent);
		text.append("\nparent id: ");
		text.append(parentId);
		if (isSource)
		{
			text.append("\nnode is input node");
		}
		text.append("\nchannel partner nodes: ");
		if ((programMatrix != null) && (programMatrix.length > 0))
		{
			text.append(programMatrix[0]);
			for (i = 1; i < programMatrix.length; i++)
			{
				text.append(", ");
				text.append(programMatrix[i]);
			}
		}
		text.append("\ndeficit: ");
		if ((deficit != null) && (deficit.length > 0))
		{
			text.append(deficit[0]);
			for (i = 1; i < deficit.length; i++)
			{
				text.append(", ");
				text.append(deficit[i]);
			}
		}
		text.append("\nsend-acknowledgment buffer sizes: ");
		if ((sendAcknowledgement != null) && (sendAcknowledgement.length > 0))
		{
			text.append(sendAcknowledgement[0]);
			for (i = 1; i < sendAcknowledgement.length; i++)
			{
				text.append(", ");
				text.append(sendAcknowledgement[i]);
			}
		}
		return text.toString();
	}
}
